import argparse
import json
import logging
import os
import yaml
import requests
import urllib3
import clients as app_clients

urllib3.disable_warnings()
logger = logging.getLogger("generate_service_connector")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)


def main(environment=None, x_auth_token=None, green_only=False):
    CURRENT_FILE_PATH = os.path.dirname(os.path.abspath(__file__))
    environment_choices = ['localdev', 'dev08', 'dev01', 'test03', 'eap01', 'prod01', 'prod02', 'prod04', 'prod05',
                           'prod06', 'prodnrt', 'prodsyd']
    if environment is not None:
        if environment not in environment_choices:
            raise ValueError('Incorrect environment choice')
    parser = argparse.ArgumentParser(description="Generate service connectors and associate grants")
    parser.add_argument("-t", "--auth-token", required=False)
    parser.add_argument('-e', '--environment', required=False,
                        choices=environment_choices)
    parser.add_argument("-g", "--green-only", help="Run migration hitting only green service", action="store_true")

    args = parser.parse_known_args()[0]

    env = args.environment if environment is None else environment
    if env not in environment_choices:
        raise ValueError('Incorrect environment choice')
    env_urls = yaml.load(open("{}/url_configs.yml".format(CURRENT_FILE_PATH)), Loader=yaml.BaseLoader)[env]
    dashboard_url = env_urls['dashboard']
    es_url = env_urls['elasticsearch']

    token = args.auth_token if x_auth_token is None else x_auth_token
    psc_client = app_clients.PscConnectorClient(dashboard_url, token)
    esclient = app_clients.EsClient(es_url)
    psc_client_set_response = psc_client.setData()
    if psc_client_set_response == 403:
        return
    esclient.setData()

    for filename in os.listdir("{}/connectors/orgs/1".format(CURRENT_FILE_PATH)):
        if filename.endswith(".json"):
            try:
                f = open("{}/connectors/orgs/1/{}".format(CURRENT_FILE_PATH, filename), 'r')
            except IOError:
                print("ERROR: An IO error occured when reading from file roles/{}".format(filename))
            else:
                text = f.read()
                service_connector_json = json.loads(text)
                service_connector_json['role_urn'] = service_connector_json['role_urn'].format(
                    org_key=psc_client.org_key)

                service_connector_name = service_connector_json['connector_name']
                query_response = esclient.queryConnector(service_connector_name.lower()).json()
                connector_id = get_connector_id(query_response, service_connector_name)
                service_connector_urn = service_connector_json['role_urn']
                if connector_id != None:
                    logger.info("----------------------------------------------------------------------------------")
                    logger.info("Service connector %s already exists, no need to recreate", service_connector_name)

                    # Verify Service Connector has correct Service Role
                    principle_urn = "psc:cnn:{}:{}".format(psc_client.org_key, connector_id)
                    org_urn = "psc:org:{}".format(psc_client.org_key)
                    grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)
                    if green_only:
                        grants_url += "?access_green_only=true"
                    grant_response = psc_client.get(path=grants_url)

                    if grant_response.ok:
                        if service_connector_urn not in grant_response.json()['roles']:
                            logger.info("Service Connector %s assigned incorrect service role %s. Updating to correct "
                                        "role %s" % (service_connector_name, str(grant_response.json()['roles']),
                                                     service_connector_urn))
                            grant_update_request = grant_response.json()
                            grant_update_request['roles'] = [service_connector_urn]

                            grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)
                            if green_only:
                                grants_url += "?access_green_only=true"
                            # Update Service Connector Role
                            grant_update_response = psc_client.put(
                                path=grants_url,
                                payload=grant_update_request)
                            grant_update_response.raise_for_status()

                    else:
                        logger.info("Grants response for %s in org %s is %s", principle_urn, org_urn, grant_response.status_code)
                    continue

                try:
                    logger.info("----------------------------------------------------------------------------------")
                    logger.info("Creating service connector with name '%s'" % service_connector_name)
                    create_connector_response = psc_client.post(
                        path="/appservices/v5/orgs/{}/connectors".format(psc_client.org_id),
                        payload={
                            "name": service_connector_json['connector_name'],
                            "description": service_connector_json['description'],
                            "authorizedIpAddresses": None,
                            "connectorType": "CUSTOM",
                            "changedBy": "secopsRoleManager@carbonblack.com"
                        }
                    )

                    create_connector_response.raise_for_status()
                    connector_id = create_connector_response.json()["connectorId"]

                    logger.info(
                        "Connector '%s' was created with connector ID '%s'" % (service_connector_name, connector_id))

                    principle_urn = "psc:cnn:{}:{}".format(psc_client.org_key, connector_id)
                    org_urn = "psc:org:{}".format(psc_client.org_key)

                    logger.info("Granting principal '%s' role '%s'" % (principle_urn, service_connector_urn))

                    grants_url = "/access/v1/grants/{}/org-ref/{}".format(principle_urn, org_urn)
                    if green_only:
                        grants_url += "?access_green_only=true"
                    grant_response = psc_client.post(
                        path=grants_url,
                        payload={
                            "principal": principle_urn,
                            "expires": None,
                            "revoked": False,
                            "roles": [service_connector_urn],
                            "org_ref": org_urn
                        }
                    )
                    grant_response.raise_for_status()

                    logger.info("Grant principal '%s' role '%s' successful" % (principle_urn, service_connector_urn))

                except requests.HTTPError as ex:
                    logger.error("Request failed (json=%s)" % ex.response.__dict__, exc_info=ex)
    logger.info("Done Migrating Service Connectors")


def get_connector_id(query_response, connector_name):
    if query_response['hits']['hits']:
        for hit in query_response['hits']['hits']:
            if hit['_source']['name'] == connector_name and len(hit['_source']['connectorId']) != 64:
                connector_id = hit['_source']['connectorId']
                return connector_id
    return None


if __name__ == "__main__":
    main()
