import psycopg2
import boto3
import simplejson as json
import argparse

from boto3.dynamodb.conditions import Attr

def main():
    source_env_choices = ['localdev', 'dev08', 'dev01', 'test03', 'eap01', 'prod01', 'prod02', 'prod04',
                           'prod05', 'prod06', 'prodnrt', 'prodsyd']
    dest_env_choices = ['localdev', 'hub02', 'hub04', 'hub-prod-us', 'hub-prod-eu', 'hub-prod-ap', 'hub-prod-ap2']
    migration_parser = argparse.ArgumentParser()
    migration_parser.add_argument('-s', '--src', required=True, choices=source_env_choices)
    migration_parser.add_argument('-d', '--dest', required=True, choices=dest_env_choices)
    args = migration_parser.parse_args()
    src_env = args.src
    dest_env = args.dest

    # postgres config
    postgres_cursor = create_postgres_client(src_env)

    #dynamoDB config
    grants_table_name = '{}_access_grants'.format(dest_env)
    #used for scan
    dynamodb_client = create_dynamodb_client(dest_env)
    #used for put_item to take advantage of optimistic locking
    dynamodb_resource = create_dynamodb_resource(dest_env)
    dynamodb_table = dynamodb_resource.Table(grants_table_name)

    # Get all users from table
    postgres_cursor.execute("SELECT login_id, login_name FROM admin_login;")
    login_results_list = postgres_cursor.fetchall()

    # Create map of principal to postgres email with join on admin_login, admin_registration, and organization
    postgres_cursor.execute("select 'psc:user:' || org.org_key || ':' || login.login_id as urn, login.login_name from admin_login login join admin_registration reg on login.login_name = reg.email_id join organization org on reg.org_id = org.org_id;")
    urn_email_list = postgres_cursor.fetchall()

    postgres_cursor.close()

    login_dict = {}
    for entry in login_results_list:
        login_id = entry[0]
        email = entry[1]
        login_dict[login_id] = email

    urn_email_dict = {}
    for entry in urn_email_list:
        urn = entry[0]
        email = entry[1]
        urn_email_dict[urn] = email

    scan_results = dynamodb_client.scan(TableName=grants_table_name)
    combined_results = scan_results['Items']
    while 'LastEvaluatedKey' in scan_results:
        scan_results = dynamodb_client.scan(TableName=grants_table_name, ExclusiveStartKey=scan_results['LastEvaluatedKey'])
        combined_results.extend(scan_results['Items'])

    no_email_list = []
    conditional_errors = 0
    for item in combined_results:
        principal_urn = item['principal']['S']
        principal_id = get_principal_id(principal_urn)

        # Only populate the field if it is not already (for reruns)
        if 'principal_name' not in item:
            try:
                if principal_urn.startswith('psc:cnn'):
                    item['principal_name'] = {'S': principal_id}
                else:
                    item['principal_name'] = {'S': login_dict[int(principal_id)]}
                dynamodb_put(dynamodb_table, item)
            except dynamodb_table.meta.client.exceptions.ConditionalCheckFailedException:
                # Script needs to be rerun to migrate this entity
                conditional_errors += 1
            except KeyError as e:
                # There was no email found for this user
                no_email_list.append(item['principal'])
            except Exception as e:
                # Unexpected
                print(e)
        else:
            # We need to validate that this principal name is correct, no collisions
            try:
                if principal_urn.startswith('psc:user') and principal_urn in urn_email_dict.keys():
                    current_principal_name = item['principal_name']['S']
                    real_principal_name = urn_email_dict[principal_urn]
                    if current_principal_name != real_principal_name:
                        print("Incorrect email found for principal %s with email %s, updating Principal name....." % (principal_urn, current_principal_name))
                        item['principal_name'] = {'S': str(real_principal_name)}
                        print("Old principal name %s, new principal name %s" % (current_principal_name, real_principal_name))
                        dynamodb_put(dynamodb_table, item)
            except dynamodb_table.meta.client.exceptions.ConditionalCheckFailedException:
                # Script needs to be rerun to migrate this entity
                conditional_errors += 1
            except KeyError as e:
                # There was no email found for this user
                no_email_list.append(item['principal'])
            except Exception as e:
                # Unexpected
                print(e)


    print("Grants without emails: {}".format(no_email_list))
    print("Conditional failure count: {}".format(conditional_errors))
        # print(item)


def get_principal_id(urn):
    parts = urn.split(':')
    return parts[3]

def get_principal_org_key(urn):
    parts = urn.split(':')
    return parts[2]


def dynamodb_put(dynamodb_table, item):
    # The item needs to be reformatted to work with the Boto3 resource APIs. Client API does not support any
    # way of using version locking.
    item_version = None
    reformatted_item = {}
    for key in item.keys():
        dynamo_val = item[key]
        inner_key = list(dynamo_val)[0]
        inner_val = dynamo_val[inner_key]
        if key == 'version' or key == 'revoked' or key == 'created_at' or key == 'updated_at':
            # Use the correct data dype so DynamoDB picks it up correctly ("N")
            inner_val = int(inner_val)
            if key == 'version':
                item_version = inner_val
                # Bump the version for optimistic locking
                inner_val += 1
        elif key == 'roles':
            # Use the correct data dype so DynamoDB picks it up correctly ("SS")
            inner_val = set(inner_val)
        reformatted_item[key] = inner_val

    dynamodb_table.put_item(
        Item=reformatted_item,
        # support for optimistic locking
        ConditionExpression=Attr('version').eq(item_version)
    )


def create_dynamodb_client(env):
    if env == 'localdev':
        dynamodb_client = boto3.client('dynamodb', region_name='us-east-1', endpoint_url='https://localhost:4599', verify=False)
    else:
        dynamodb_client = boto3.client('dynamodb')
    return dynamodb_client


def create_dynamodb_resource(env):
    if env == 'localdev':
        dynamodb_client = boto3.resource('dynamodb', region_name='us-east-1', endpoint_url='https://localhost:4599', verify=False)
    else:
        dynamodb_client = boto3.resource('dynamodb')
    return dynamodb_client


def create_postgres_client(env):
    if env =='localdev':
        conn_string = "port='5432' user='localdev' password='localdev' host='localhost'"
    else:
        with open('rds.json') as postgres_json:
            data = json.load(postgres_json)
            hostname = data['rds_hostname']
            username = data['rds_username']
            password = data['rds_password']
        conn_string = "dbname='{}' port='5432' user={} password='{}' host='{}'".format(env, username,
                                                                                       password, hostname)
    conn = psycopg2.connect(conn_string)
    cur = conn.cursor()
    return cur


if __name__ == "__main__":
    main()

