import argparse
import urllib3
import logging
import os

import migrate_secops_role_manager as migrate_secops
from migration import migrate_v2_entities as migrate_roles_and_permissions
import add_service_connectors as migrate_service_connectors
import remove_manager_permission as remove_secops_role_manager
import custom_role_subrole_migration as migrate_subroles_for_custom_roles

urllib3.disable_warnings()
logger = logging.getLogger("Manager for migrate all v2 RBAC")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)

def main():
    migration_parser = argparse.ArgumentParser(description="Manager for migrating roles and permissions")
    environment_choices = ['localdev', 'dev08', 'dev01', 'test03', 'eap01', 'prod01', 'prod02', 'prod04', 'prod05', 'prod06', 'prodnrt', 'prodsyd']
    migration_parser.add_argument('-e', '--environment', required=True,
                        choices=environment_choices)
    migration_parser.add_argument('-csp', "--migrate-csp-entities", required=False, action="store_true", default=False, help="Run migration of CSP entities too.")
    migration_parser.add_argument("-g", "--green-only", help="Run migration hitting only green service", action="store_true")
    migration_parser.add_argument("-s", "--service-connectors", help="Run only add service connectors", action="store_true")
    migration_parser.add_argument("-m", "--migrate-only", help="Run only roles and permission migration", action="store_true")
    migration_parser.add_argument("-p", "--migration-plan", help="Only print the migration plan", action="store_true")
    migration_parser.add_argument("-csu", "--custom-subroles-updates", help="Update custom roles with new subroles", action="store_true")

    args = migration_parser.parse_args()
    env = args.environment
    only_service_connectors = args.service_connectors
    only_role_and_permissions = args.migrate_only
    green_only = args.green_only
    plan = args.migration_plan
    custom_subroles_updates = args.custom_subroles_updates
    migrate_csp_entities = args.migrate_csp_entities

    if green_only:
        logger.info("GREEN ONLY IS ENABLED")
    
    # prd org id: bd4b3760-eb57-41b2-b638-4257a830c7dd
    # stg org id: 9a9277a0-01c2-4a93-98fb-4392975efa56
    # Both having:
    # serviceName: "VMware Carbon Black Cloud"
    # serviceDisplayName: "VMware Carbon Black Cloud"
    csp_struct = {}
    if migrate_csp_entities:
        csp_env_key = "STG"
        if env.startswith("prod") or env.startswith("eap"):
            csp_env_key = "PROD"
            csp_struct["url"] = "https://console.cloud.vmware.com"
            csp_struct["service_definition_id"] = "78962d09-1571-4db1-b9a0-1bdb69045252"
        else:
            csp_struct["url"] = "https://console-stg.cloud.vmware.com"
            csp_struct["service_definition_id"] = "70539df3-4c23-41a3-b214-26c96f261b13"
        # Expects both env vars to be set (usually by Gitlab) 
        csp_struct["client_id"] = os.environ[f"CSP_{csp_env_key}_CLIENT_ID"]
        csp_struct["secret_id"] = os.environ[f"CSP_{csp_env_key}_SECRET_ID"]
    
    # Create or Read Secops Role Manager and retrieve auth-token and org key
    secop_manager_connector_dict = migrate_secops.main(environment=env, green_only=green_only)

    auth_token = secop_manager_connector_dict['X-Auth-Token']
    org_key = secop_manager_connector_dict['org-key']
    if not only_service_connectors or only_role_and_permissions:
        logger.info("------------------ADDING ROLES AND PERMISSIONS------------------")
        # Migrate all roles and permissions with the exported auth token of the secops role manager
        migrate_roles_and_permissions.main(environment=env, x_auth_token=auth_token, org_key=org_key, all=True, green_only=green_only, plan=plan, csp_struct=csp_struct)

    if not only_role_and_permissions or only_service_connectors:
        logger.info("------------------ADDING SERVICE CONNECTORS------------------")
        # Migrate all service connectors
        migrate_service_connectors.main(environment=env, x_auth_token=auth_token, green_only=green_only)

    # If there are custom roles that need to be modified to add subroles
    if custom_subroles_updates:
        logger.info("------------------ADDING SUBROLES TO CUSTOM ROLES------------------")
        try:
            migrate_subroles_for_custom_roles.main(plan=plan, env=env, x_auth_token=auth_token)
        except Exception as e:
            logger.exception(e)

    # Remove Secops Role Manager Permissions
    remove_secops_role_manager.main(environment=env)

if __name__ == "__main__":
    main()
