import boto3
import requests
from typing import List
from datetime import date
import json
import urllib3
import csv
import os
from collections import namedtuple
import logging

urllib3.disable_warnings()
logger = logging.getLogger("Secops-Grant-Export")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s %(levelname)s - %(message)s")
handler_console = logging.StreamHandler()
handler_console.setFormatter(formatter)
logger.addHandler(handler_console)

class CBCEnvironment:
    """
    Wrapper that represents an environment in the Carbon Black Cloud ecosystem
    """
    def __init__(self, env: str, host: str, gateway_url: str):
        """
        :param env: name of the CBC environment (e.g. 'dev01')
        """
        self.env = env
        self.host = host
        self.gateway_url = gateway_url

    @staticmethod
    def from_dict(d):
        return namedtuple("CBCEnvironment", d.keys())(*d.values())


def split_keys(api_keys: str, api_keys_delimeter=":", env_delimeter="@"):
    keys = dict(key.split(api_keys_delimeter) for key in api_keys.split(env_delimeter))
    return {k.lower(): v for k, v in keys.items()}

"""
In format hub-dev-us
"""
HUB_NAME = os.environ.get('HUB_NAME', '')

"""
 In format [{"env":"local", "host": "psc.gateway.net", "gateway_url": "https://psc:gateway.net"}]
"""
CBC_ENVS = list(map(lambda o: CBCEnvironment.from_dict(o), json.loads(os.environ.get('CBC_ENVS', '{}'))))

"""
In the following format
dev01:secret_dev01/key_dev01@test03:secret_test03/key_test03@prod05:secret_prod05/key_prod05"""
CBC_AUTH_TOKENS = split_keys(api_keys=os.environ.get('CBC_AUTH_TOKENS', '{}').rstrip())


class Grant:
    def __init__(self, urn: str,  org_ref: str, principal_name: str, roles: List[str]):
        self.urn = urn
        self.org_ref = org_ref
        self.principal_name = principal_name
        self.roles = roles

    def get_org_key(self):
        """
        psc:org:ABC123
        :return: org key in the org_ref, in the above example will return ABC123
        """
        parts = [x for x in self.org_ref.split(":")]
        return parts[2]


class Organization:
    def __init__(self, orgKey: str, name: str, id: int):
        self.orgKey = orgKey
        self.name = name
        self.id = id


class BaseClient:
    def __init__(self, base_url, auth_token, host):
        self.base_url = base_url
        self.session = None
        self.headers = None
        self.auth_token = auth_token
        self.host = host
        self.headers = {
            "Content-Type": "application/json",
            "X-Auth-Token": self.auth_token,
            "Host": host
        }
        self.session = requests.Session()

    def find_all_orgs(self) -> List[Organization]:
        """
        Finds all organizations in CB Defense.

        :return: list of orgs
        """
        csr_url = '{base_url}{svc_path}'.format(base_url=self.base_url, svc_path="/adminServices/v5/orgs")
        logging.info('Making request to CSR Admin Services to find all orgs...')
        logging.info('URL = {}'.format(csr_url))
        try:
            r = requests.get(csr_url, headers=self.headers, verify=False)
            if 400 <= r.status_code < 600:
                raise Exception('Failed to find all orgs. Request returned {} response code. Body = {}'
                                .format(r.status_code, str(r.content)))

            organizations = json.loads(r.content)['organizations']

            logging.info('Found {} orgs.'.format(len(organizations)))

            return list(map(lambda o: Organization(
                orgKey=o['orgKey'],
                name=o['name'],
                id=o['id']
            ), organizations))
        except Exception as e:
            logging.error('Error occurred while making a request to CSR gateway.')
            logging.exception(e)
            raise e


class DynamoDbClient:
    def __init__(self, environment):
        self.environment = environment
        self.dynamo_client = None
        if self.environment == 'localdev':
            self.dynamoClient = boto3.client('dynamodb', region_name='us-east-1', endpoint_url='https://access-mockaws:4569', verify=False)
        else:
            self.dynamoClient = boto3.client('dynamodb')

    def scan(self, table_name) -> List[Grant]:
        response = self.dynamoClient.scan(TableName=table_name)
        full_response = None

        # Needed since table can be much larger than boto can grab in one scan
        if 'LastEvaluatedKey' in response:
            full_response = response['Items']
            while 'LastEvaluatedKey' in response:
                response = self.dynamoClient.scan(TableName=table_name, ExclusiveStartKey=response['LastEvaluatedKey'])
                full_response.extend(response['Items'])
                scan_num = len(full_response)
                logger.info(f"Scanning {table_name} table... total items scanned: {scan_num}", )
            logger.info(f"Done scanning {table_name}.")

        full_response = full_response if full_response is not None else response
        grants_list = []
        for g in full_response:
            principal_name = g['principal_name']['S'] if 'principal_name' in g else ""
            roles_or_grant = g['roles']['SS'] if 'roles' in g else g['profiles']['L']
            grants_list.append(
                Grant(
                    org_ref=g['org_ref']['S'],
                    principal_name=principal_name,
                    urn=g['principal']['S'],
                    roles=roles_or_grant
                )
            )

        return grants_list


def throw_api_key_exception(config_name):
    raise Exception(f'There are API keys missing for {config_name}. '
                    'Please make sure that there is an API key for each configured env in CBC_ENVS.')

def log_api_key_warning(config_name):
    logging.warning(f"There are more API keys configured for {config_name} than environments "
                    "configured in CBC_ENVS. Only what is included in CBC_ENVS will run for this job.")

def verify_run_config():
    """ Verifies configuration required at runtime
    In format [{"env":"local"}, {"env":"dev01"}, {"env":"test03"}]
    """
    cbc_envs = list(map(lambda e: e.env.lower(), CBC_ENVS))
    auth_tokens = CBC_AUTH_TOKENS.keys()

    if len(cbc_envs) == 0:
        logging.warning("Cannot run the job if there are no cbc_envs configured.")
        raise Exception("exiting...")
    else:
        if not all(item in cbc_envs for item in auth_tokens):
            throw_api_key_exception(config_name='CBC_AUTH_TOKENS')

        if len(auth_tokens) > len(cbc_envs):
            log_api_key_warning(config_name='CBC_AUTH_TOKENS')


def main():
    verify_run_config()
    today = date.today().strftime("%b-%d-%Y")
    dynamo_client = DynamoDbClient(environment=HUB_NAME)
    # Get all grants once in the hub-prod-us table
    all_grants = dynamo_client.scan(f"{HUB_NAME}_access_grants")
    with open(f'/tmp/{HUB_NAME}-secops-grant-export-{today}.csv', 'wt') as myfile:
        writer = csv.writer(myfile, delimiter=',', quoting=csv.QUOTE_ALL)
        writer.writerow(['env', 'orgId', 'orgKey', 'orgName', 'principalName', 'roles'])

        for cbc_env in CBC_ENVS:
            env = cbc_env.env
            logger.info(f"Running in {env}...")
            host = cbc_env.host
            url = cbc_env.gateway_url

            auth_token = CBC_AUTH_TOKENS[env]
            base_client = BaseClient(base_url=url, auth_token=auth_token, host=host)

            all_orgs = base_client.find_all_orgs()

            # Create map of orgKey to org for easy parsing
            orgkey_dict = {o.orgKey: o for o in all_orgs}

            for grant in all_grants:
                org_key = grant.get_org_key()
                if org_key in orgkey_dict:
                    # Write to the csv
                    org_id = orgkey_dict[grant.get_org_key()].id
                    org_name = orgkey_dict[grant.get_org_key()].name
                    writer.writerow([env, org_id, org_key, org_name, grant.principal_name,
                                     str(grant.roles).replace("\'", "")])


if __name__ == "__main__":
    main()